from datetime import date


__version__ = '1.1.1.dev3'
"""The current version of the TurboCtl project."""


copyright = f'2019-{date.today().year} University of Helsinki Fusor Team'
"""Automatically updating copyright information."""
